-------------------------------------------------------------------------------
-- Title      : DIOT Demo Diagnostics
-- Project    : Distributed I/O Tier
-------------------------------------------------------------------------------
-- File       : diot_diag.vhd
-- Author     : Greg Daniluk <grzegorz.daniluk@cern.ch>
-- Company    : CERN (BE-CO-HT)
-- Standard   : VHDL
-------------------------------------------------------------------------------
-- Description:
-- This module exports basic monitoring information provided by DIOT
-- demonstrator crates.
-------------------------------------------------------------------------------
--
-- Copyright (c) 2018 CERN
--
-- Copyright and related rights are licensed under the Solderpad Hardware
-- License, Version 0.51 (the “License”) (which enables you, at your option,
-- to treat this file as licensed under the Apache License 2.0); you may not
-- use this file except in compliance with the License. You may obtain a copy
-- of the License at http://solderpad.org/licenses/SHL-0.51.
-- Unless required by applicable law or agreed to in writing, software,
-- hardware and materials distributed under this License is distributed on an
-- “AS IS” BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
-- or implied. See the License for the specific language governing permissions
-- and limitations under the License.
--
-------------------------------------------------------------------------------
--
-- FCM2 commands layout:
-- read i2c address: 0x39
-- cmd 0: fan  status (return 1 Byte)
-- cmd 1: temp status (return 1 Byte)
-- cmd 2: volt status (return 1 Byte)
-- cmd 3: fans speed  (return 12 Bytes) - 2B per fan
-- cmd 4: temperature (return 6 Bytes)  - 1B per sensor
--
-------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use work.diag_pkg.all;

entity diot_diag is
  port (
    rst_n_i : in  std_logic;
    clk_i   : in  std_logic;

    scl_o   : out std_logic;
    scl_oen : out std_logic;
    scl_i   : in  std_logic;
    sda_o   : out std_logic;
    sda_oen : out std_logic;
    sda_i   : in  std_logic;

    diags_o : out t_diag_out
  );
end diot_diag;

architecture rtl of diot_diag is

  constant c_FCM_ADR_W : std_logic_vector(7 downto 0) := x"38";
  constant c_FCM_ADR_R : std_logic_vector(7 downto 0) := x"39";
  -- how many commands should be executed on FCM2, we're insterested in values
  -- from commands 0 - 4.
  constant c_CMDS        : integer := 5;
  constant c_WAIT_CYCLES : integer := 30000;
  type t_state is (IDLE, START, SEND_CMD, RP_START, READ, RD_LAST, PAUSE);

  -- signals for i2c master component
  signal i2c_en      : std_logic;
  signal i2c_start   : std_logic;
  signal i2c_stop    : std_logic;
  signal i2c_rd      : std_logic;
  signal i2c_wr      : std_logic;
  signal i2c_txack   : std_logic;
  signal i2c_busy    : std_logic;
  signal i2c_lost    : std_logic;
  signal i2c_txbyte  : std_logic_vector(7 downto 0);
  signal i2c_rxbyte  : std_logic_vector(7 downto 0);
  signal i2c_done    : std_logic;
  signal i2c_rxack_n : std_logic;
  signal i2c_rxack   : std_logic;

  signal rst       : std_logic;
  signal state     : t_state;
  signal cmd       : unsigned(7 downto 0);
  signal cnt       : unsigned(15 downto 0);
  signal fan_buf   : t_diag_buf(0 to 11);
  signal last_byte : std_logic;

begin

  rst       <= not rst_n_i;
  i2c_rxack <= not i2c_rxack_n;

  cmp_i2c_master: entity work.i2c_master_byte_ctrl
  port map (
    clk      => clk_i,
    rst      => rst,
    nReset   => rst_n_i,
    ena      => i2c_en,
    clk_cnt  => x"0080",
    start    => i2c_start,
    stop     => i2c_stop,
    read     => i2c_rd,
    write    => i2c_wr,
    ack_in   => i2c_txack,

    i2c_busy => i2c_busy,
    i2c_al   => i2c_lost,
    din      => i2c_txbyte,
    cmd_ack  => i2c_done,
    ack_out  => i2c_rxack_n,
    dout     => i2c_rxbyte,

    scl_i    => scl_i,
    scl_o    => scl_o,
    scl_oen  => scl_oen,
    sda_i    => sda_i,
    sda_o    => sda_o,
    sda_oen  => sda_oen
  );

  process(clk_i)
  begin
    if rising_edge(clk_i) then
      if rst_n_i = '0' then
        i2c_en     <= '0';
        i2c_start  <= '0';
        i2c_stop   <= '0';
        i2c_rd     <= '0';
        i2c_wr     <= '0';
        i2c_txack  <= '0';
        i2c_txbyte <= (others=>'0');
        cmd        <= (others=>'0');
        cnt        <= (others=>'0');
        fan_buf    <= (others=>(others=>'0'));
        diags_o    <= c_DIAGS_DUMMY;
        state      <= IDLE;
      else
        case state is
          ---------------------------------------
          when IDLE =>
            i2c_en     <= '1';
            i2c_start  <= '0';
            i2c_stop   <= '0';
            i2c_rd     <= '0';
            i2c_wr     <= '0';
            i2c_txack  <= '0';
            i2c_txbyte <= (others=>'0');
            cmd        <= (others=>'0');
            cnt        <= (others=>'0');

            state <= START;

          ---------------------------------------
          when START =>
            -- send I2C start with FCM2 address for writing
            i2c_en     <= '1';
            i2c_start  <= '1';
            i2c_stop   <= '0';
            i2c_rd     <= '0';
            i2c_wr     <= '1';
            i2c_txack  <= '0';
            i2c_txbyte <= c_FCM_ADR_W;
            cnt        <= (others=>'0');

            if (i2c_done = '1') then
              i2c_start <= '0';
              i2c_wr    <= '0';
            end if;
            if (i2c_done = '1' and i2c_rxack = '1') then
              state     <= SEND_CMD;
            elsif (i2c_done = '1') then
              i2c_stop  <= '1';
              state     <= PAUSE;
            end if;

          ---------------------------------------
          when SEND_CMD =>
            -- write command index to FCM2
            i2c_en     <= '1';
            i2c_start  <= '0';
            i2c_stop   <= '1';
            i2c_rd     <= '0';
            i2c_wr     <= '1';
            i2c_txack  <= '0';
            i2c_txbyte <= std_logic_vector(cmd);
            cnt        <= (others=>'0');

            if (i2c_done = '1') then
              i2c_stop  <= '0';
              i2c_wr    <= '0';
            end if;

            if (i2c_done = '1' and i2c_rxack = '1') then
              state     <= RP_START;
            elsif (i2c_done = '1') then
              state     <= PAUSE;
            end if;
          ---------------------------------------
          when RP_START =>
            -- repeat start to initialize data readout, send FCM2 address for
            -- reading
            i2c_en     <= '1';
            i2c_start  <= '1';
            i2c_stop   <= '0';
            i2c_rd     <= '0';
            i2c_wr     <= '1';
            i2c_txack  <= '0';
            i2c_txbyte <= c_FCM_ADR_R;
            cnt        <= (others=>'0');

            if (i2c_done = '1') then
              i2c_start <= '0';
              i2c_wr    <= '0';
            end if;

            if (i2c_done = '1' and i2c_rxack = '1' and cmd < c_CMD_FAN_SPEED) then
              -- for cmd 0, 1, 2 there is only 1 byte to read
              state     <= RD_LAST;
            elsif (i2c_done = '1' and i2c_rxack = '1') then
              -- for cmd 3, 4 there are multiple bytes to read
              state     <= READ;
            elsif (i2c_done = '1') then
              state     <= PAUSE;
            end if;
          ---------------------------------------
          when READ =>
            -- reading result of requested command
            i2c_en     <= '1';
            i2c_start  <= '0';
            i2c_stop   <= '0';
            i2c_rd     <= '1';
            i2c_wr     <= '0';
            i2c_txack  <= '0';
            i2c_txbyte <= (others=>'0');

            diags_o.fans_speed(0 to 5) <= fan_buf(0 to 5);
            -- save read byte to appropriate place
            if (i2c_done = '1' and cmd = c_CMD_FAN_SPEED) then
              fan_buf(to_integer(cnt))  <= i2c_rxbyte;
            elsif (i2c_done = '1' and cmd = c_CMD_TEMP) then
              diags_o.temps(to_integer(cnt)) <= i2c_rxbyte;
            end if;

            if (i2c_done = '1' and last_byte = '1') then
              -- next byte will be the last one, needs different handling
              i2c_rd   <= '0';
              cnt      <= cnt + 1;
              state    <= RD_LAST;
            elsif (i2c_done = '1') then
              -- continue reading bytes
              i2c_rd   <= '0';
              cnt      <= cnt + 1;
            end if;
          ---------------------------------------
          when RD_LAST =>
            -- reading result of requested command
            i2c_en     <= '1';
            i2c_start  <= '0';
            i2c_stop   <= '1';
            i2c_rd     <= '1';
            i2c_wr     <= '0';
            i2c_txack  <= '1';
            i2c_txbyte <= (others=>'0');

            diags_o.fans_speed(0 to 5) <= fan_buf(0 to 5);
            -- save read byte to appropriate place
            if (i2c_done = '1' and cmd = c_CMD_FAN_STAT) then
              diags_o.fan_status        <= i2c_rxbyte;

            elsif (i2c_done = '1' and cmd = c_CMD_TEMP_STAT) then
              diags_o.temp_status       <= i2c_rxbyte;

            elsif (i2c_done = '1' and cmd = c_CMD_VOLT_STAT) then
              diags_o.volt_status       <= i2c_rxbyte;

            elsif (i2c_done = '1' and cmd = c_CMD_FAN_SPEED) then
              fan_buf(to_integer(cnt))  <= i2c_rxbyte;

            elsif (i2c_done = '1' and cmd = c_CMD_TEMP) then
              diags_o.temps(to_integer(cnt)) <= i2c_rxbyte;
            end if;

            if (i2c_done = '1' and cmd < c_CMD_TEMP) then
              -- execute next command
              i2c_stop <= '0';
              i2c_rd   <= '0';
              cnt      <= (others=>'0');
              cmd      <= cmd + 1;
              state    <= START;
            elsif (i2c_done = '1' and cmd = c_CMD_TEMP) then
              -- all info is read, pause before next readout cycle starts
              i2c_stop <= '0';
              i2c_rd   <= '0';
              cnt      <= (others=>'0');
              state    <= PAUSE;
            end if;

          ---------------------------------------
          when PAUSE =>
            -- wait before starting next monitoring readout cycle
            i2c_en     <= '0';
            i2c_start  <= '0';
            i2c_rd     <= '0';
            i2c_wr     <= '0';
            i2c_txack  <= '0';
            i2c_txbyte <= (others=>'0');
            cmd        <= (others=>'0');
            cnt        <= cnt + 1;

            if (i2c_done = '1') then
              i2c_stop <= '0';
            end if;

            if (cnt = c_WAIT_CYCLES) then
              state <= IDLE;
            end if;
          ---------------------------------------
          when others =>
            i2c_en     <= '0';
            i2c_start  <= '0';
            i2c_stop   <= '0';
            i2c_rd     <= '0';
            i2c_wr     <= '0';
            i2c_txack  <= '0';
            i2c_txbyte <= (others=>'0');
            cmd        <= (others=>'0');
            cnt        <= (others=>'0');

            state <= IDLE;

        end case;
      end if;
    end if;
  end process;

  -- signal when we're reading one but last byte for multi-byte commands
  last_byte <= '1' when (cmd = c_CMD_FAN_SPEED and cnt = fan_buf'high - 1) else
               '1' when (cmd = c_CMD_TEMP and cnt = diags_o.temps'high - 1) else
               '0';

end rtl;
