/*
 * Copyright (C) 2018 CERN
 * Author: Maciej Suminski <maciej.suminski@cern.ch>
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <QtWidgets/QPushButton>
#include <QtWidgets/QGroupBox>
#include <QBoxLayout>
#include <QLabel>
#include <QStatusBar>
#include "controller.h"

// Register custom type for Qt signals
static struct registerMetatypes {
    registerMetatypes()
    {
        qRegisterMetaType<nanoFIP::DataStatusMap>("nanoFIP::DataStatusMap");
        qRegisterMetaType<mstrfip_data_status>("mstrfip_data_status");
    }
} _registerMetatypes;


DiotHardware::DiotHardware(int address, int slots_nr)
{
    // Configure MasterFIP
    mfip = MasterFIP::OpenByLUN(0);

    // Hardware and software config for MasterFIP
    memset(&hw_cfg, 0, sizeof(mstrfip_hw_cfg));
    hw_cfg.enable_ext_trig = 0;
    hw_cfg.enable_int_trig = 0;
    hw_cfg.enable_ext_trig_term = 0;
    hw_cfg.turn_around_ustime = 14;         // 14 us corresponds to 1 Mbit/s

    memset(&sw_cfg, 0, sizeof(mstrfip_sw_cfg));
    hw_cfg.enable_ext_trig = 0;
    sw_cfg.irq_thread_prio = 50;
    sw_cfg.diag_thread_prio = 20;

    // Cycle configuration (60 ms duration)
    cycle.reset(new CycleSimple(*mfip, 60000));
    crate.reset(new DIOT_Crate(address, slots_nr));
    crate->AddVariables(cycle->GetPeriodicVarWindow());

    crate->SetConsVarCb([&](nanoFIP &nf, mstrfip_data *data) {
        emit StatusChanged(data->status, nf.GetDataStats());

        // No more updates if the data is invalid
        if(data->status != MSTRFIP_DATA_OK)
            return;

        // Report the crate state
        DIOT_Crate* crate = static_cast<DIOT_Crate*>(&nf);

        for(int i = 0; i < crate->GetCardCount(); ++i) {
            emit InputChanged(i, crate->GetSlotInputs(i));

            uint32_t inpError = crate->GetSlotInputErrs(i);
            if (inpError) {
                emit InputErrChanged(i, inpError);
            }
        }

        for(unsigned int i = 0; i < crate->GetFanCount(); ++i) {
            emit FanChanged(i, crate->GetFanStatus(i), crate->GetFanRPM(i));
        }

        for(unsigned int i = 0; i < crate->GetTempCount(); ++i) {
            emit TemperatureChanged(i, crate->GetTempStatus(i), crate->GetTemp(i));
        }

        for(unsigned int i = 0; i < crate->GetVoltages().size(); ++i) {
            emit VoltageChanged(i, crate->GetVoltStatus(crate->GetVoltEnum(i)));
        }
    } );
}


void DiotHardware::Start() {
    mfip->Configure(hw_cfg, sw_cfg);
    mfip->StartCycle(*cycle);
}


void DiotHardware::SetOutput(int slot, unsigned int state)
{
    qDebug() << "sending output update: slot " << slot << "=" << state;
    crate->SetSlotOutputs(slot, state);
}


DiotIOCard::DiotIOCard(QWidget* parent, const DiotHardware& hw, int num, int idx)
    : QFrame(parent), inputs(num, true),
        outputs(num, false), slot_nr(idx), io_count(num)
{
    QHBoxLayout *mainLayout = new QHBoxLayout();

    QGridLayout *grid = new QGridLayout();
    grid->setColumnStretch(0, 0);
    grid->setColumnStretch(1, 1);
    grid->setColumnStretch(2, 1);

    QGroupBox *groupBox = new QGroupBox(this);
    groupBox->setTitle(QString("Slot %1").arg(idx + 1));

    for(int i = 0; i < io_count; ++i) {
        QLabel* state = new QLabel(QString("%1").arg((i + 1) + idx * num), groupBox);
        state->setAlignment(Qt::AlignHCenter | Qt::AlignVCenter);
        state->setMargin(3);
        grid->addWidget(state, i, 0);
        lbl_state.push_back(state);

        QPushButton* test = new QPushButton("Test", groupBox);
        test->setFixedWidth(50);
        grid->addWidget(test, i, 1);
        QObject::connect(test, &QPushButton::clicked,
                [=](){ testBtn_click(idx, i); });

        QPushButton* clearErr = new QPushButton("", groupBox);
        clearErr->setFixedWidth(70);
        clearErr->setStyleSheet("color: red");
        // "Clear Error" button is reenabled when an error occurs
        clearErr->setEnabled(false);
        grid->addWidget(clearErr, i, 2);
        btn_clrerr.push_back(clearErr);
        QObject::connect(clearErr, &QPushButton::clicked, [=](){
                clearErr->setEnabled(false);
                clearErr->setText("");
            });
    }

    groupBox->setLayout(grid);
    mainLayout->addWidget(groupBox);
    setLayout(mainLayout);
    update_inputs();

    QObject::connect(this, &DiotIOCard::InputChanged, [=](){ update_inputs(); });
    QObject::connect(this, &DiotIOCard::OutputChanged, &hw, &DiotHardware::SetOutput);
}


void DiotIOCard::SetInput(unsigned int state) {
    if(int_to_vec(state, inputs))
        emit InputChanged(slot_nr, state);
}


void DiotIOCard::SetInputErr(unsigned int state) {
    for(unsigned int i = 0; i < btn_clrerr.size(); ++i) {
        if(state & (1 << i)) {
            btn_clrerr[i]->setEnabled(true);
            btn_clrerr[i]->setText("Clear Error");
        }
    }
}


void DiotIOCard::SetOutput(unsigned int state) {
    if(int_to_vec(state, outputs))
        emit OutputChanged(slot_nr, state);
}


void DiotIOCard::testBtn_click(int slot, int io) {
    qDebug() << "button test clicked - slot(" << slot << ") io(" << io << ")";
    outputs[io] = !outputs[io];
    emit OutputChanged(slot_nr, GetOutput());
}


void DiotIOCard::update_inputs() {
    for(int i = 0; i < io_count; ++i) {
        if(!inputs[i]) {
            lbl_state[i]->setStyleSheet("background-color: green; color: white");
        } else {
            lbl_state[i]->setStyleSheet("background-color: maroon; color: white");
        }
    }
}


unsigned int DiotIOCard::vec_to_int(const std::vector<bool>& vec) {
    unsigned int res = 0;
    int mask = 1;

    for(bool state : vec) {
        if(state)
            res |= mask;

        mask <<= 1;
    }

    return res;
}


bool DiotIOCard::int_to_vec(unsigned int integer, std::vector<bool>& vec)
{
    bool changed = false;

    for(unsigned int i = 0; i < vec.size(); ++i) {
        bool new_state = (integer & 1);

        if(vec[i] != new_state) {
            vec[i] = new_state;
            changed = true;
        }

        integer >>= 1;
    }

    return changed;
}


DiotIOController::DiotIOController(QWidget* dialog, DiotHardware& hw, int slots_count)
    : QFrame(dialog)
{
    QHBoxLayout *boxLayout = new QHBoxLayout();

    // Create widgets representing I/O slots
    for(int i = 0; i < slots_count; ++i) {
        io_slots.emplace_back(new DiotIOCard(dialog, hw, IO_PER_SLOT, i));
        boxLayout->addWidget(io_slots.back().get());
    }

    setLayout(boxLayout);

    QObject::connect(&hw, &DiotHardware::InputChanged,
            this, &DiotIOController::UpdateInput);
    QObject::connect(&hw, &DiotHardware::InputErrChanged,
            this, &DiotIOController::UpdateInputErr);
}


void DiotIOController::UpdateInput(int slot, unsigned int state)
{
    qDebug() << "received input update: slot " << slot << "=" << state;
    io_slots[slot]->SetInput(state);
}


void DiotIOController::UpdateInputErr(int slot, unsigned int state)
{
    qDebug() << "received input err update: slot " << slot << "=" << state;
    io_slots[slot]->SetInputErr(state);
}


DiotDiag::DiotDiag(QWidget* parent, const DiotHardware& hw)
    : QFrame(parent)
{
    DIOT_Crate* crate = hw.GetCrate();
    QVBoxLayout *mainLayout = new QVBoxLayout();
    QHBoxLayout *statsLayout = new QHBoxLayout();

    // Fans
    QGroupBox *fanBox = new QGroupBox(this);
    QVBoxLayout *fanLayout = new QVBoxLayout();
    for(unsigned int i = 0; i < crate->GetFanCount(); ++i) {
        auto fanWidget = new QLabel(QString("Fan %1").arg(i + 1), fanBox);
        fanLayout->addWidget(fanWidget);
        fan_lbl.push_back(fanWidget);
    }
    fanBox->setTitle("Fans");
    fanBox->setLayout(fanLayout);
    statsLayout->addWidget(fanBox);

    // Temperature sensors
    QGroupBox *tempBox = new QGroupBox(this);
    QVBoxLayout *tempLayout = new QVBoxLayout();
    for(unsigned int i = 0; i < crate->GetTempCount(); ++i) {
        auto tempWidget = new QLabel(QString("Temperature %1").arg(i + 1), tempBox);
        tempLayout->addWidget(tempWidget);
        temp_lbl.push_back(tempWidget);
    }
    tempBox->setTitle("Temperature");
    tempBox->setLayout(tempLayout);
    statsLayout->addWidget(tempBox);

    // Voltages
    QGroupBox *voltBox = new QGroupBox(this);
    QVBoxLayout *voltLayout = new QVBoxLayout();
    const auto& voltages = crate->GetVoltages();
    for(unsigned int i = 0; i < voltages.size(); ++i) {
        auto voltageWidget = new QLabel(QString::fromStdString(voltages[i].second), voltBox);
        voltLayout->addWidget(voltageWidget);
        volt_lbl.push_back(voltageWidget);
    }
    voltBox->setTitle("Voltages");
    voltBox->setLayout(voltLayout);
    statsLayout->addWidget(voltBox);

    // Data statistics
    QGroupBox *statBox = new QGroupBox(this);
    QVBoxLayout *statLayout = new QVBoxLayout();
    for(const auto& statType : hw.GetCrate()->GetDataStats()) {
        auto statWidget = new QLabel(QString("%1: 0")
                .arg(QString::fromStdString(nanoFIP::GetDataStatDesc(statType.first))));
        statLayout->addWidget(statWidget);
        stat_lbl.push_back(statWidget);
    }
    statBox->setTitle("Data statistics");
    statBox->setLayout(statLayout);
    statsLayout->addWidget(statBox);

    mainLayout->addLayout(statsLayout, 1);

    statusBar = new QStatusBar(this);
    mainLayout->addWidget(statusBar);

    setLayout(mainLayout);

    QObject::connect(&hw, &DiotHardware::FanChanged,
            this, &DiotDiag::UpdateFan);
    QObject::connect(&hw, &DiotHardware::TemperatureChanged,
            this, &DiotDiag::UpdateTemperature);
    QObject::connect(&hw, &DiotHardware::VoltageChanged,
            this, &DiotDiag::UpdateVoltage);
    QObject::connect(&hw, &DiotHardware::StatusChanged,
            this, &DiotDiag::UpdateStatus);
}


void DiotDiag::UpdateVoltage(unsigned int volt, bool state)
{
    QLabel* label = volt_lbl[volt];
    auto volt_enum = DIOT_Crate::GetVoltEnum(volt);
    qDebug() << "voltage " << DIOT_Crate::GetVoltName(volt_enum).c_str() << " state: " << state;

    if(state) {
        label->setStyleSheet("color: green");
        label->setText(QString("%1 OK").arg(
                    QString::fromStdString(DIOT_Crate::GetVoltName(volt_enum))));
    } else {
        label->setStyleSheet("color: red");
        label->setText(QString("%1 ERR").arg(
                    QString::fromStdString(DIOT_Crate::GetVoltName(volt_enum))));
    }
}


void DiotDiag::UpdateStatus(mstrfip_data_status status,
        const nanoFIP::DataStatusMap& stats)
{
    qDebug() << "status: " << status;

    switch(status)
    {
        case MSTRFIP_DATA_OK:
            statusBar->showMessage("Status: data OK");
            break;

        case MSTRFIP_DATA_FRAME_ERROR:
            statusBar->showMessage("Status: frame error");
            break;

        case MSTRFIP_DATA_PAYLOAD_ERROR:
            statusBar->showMessage("Status: payload error");
            break;

        case MSTRFIP_DATA_NOT_RECEIVED:
            statusBar->showMessage("Status: data not received");
            break;
    }

    if(status == MSTRFIP_DATA_OK) {
        statusBar->setStyleSheet("color: green");
        parentWidget()->setEnabled(true);
    }
    else {
        statusBar->setStyleSheet("color: red");
        parentWidget()->setEnabled(false);
    }

    for( const auto& stat : stats )
    {
        stat_lbl[static_cast<int>(stat.first)]->setText(QString("%1: %2")
                    .arg(QString::fromStdString(nanoFIP::GetDataStatDesc(stat.first)))
                    .arg(stat.second));
    }
}


void DiotDiag::updateLabel(std::vector<QLabel*> labels, unsigned int idx,
        bool state, const QString& desc, int val)
{
    if(idx >= labels.size())
        return; // TODO assert?

    QLabel* label = labels[idx];

    if(state) {
        label->setStyleSheet("color: green");
        label->setText(QString("%1 %2 OK: %3").arg(desc).arg(idx + 1).arg(val));
    } else {
        label->setStyleSheet("color: red");
        label->setText(QString("%1 %2 ERR: %3").arg(desc).arg(idx + 1).arg(val));
    }
}
