/*
 * Copyright (C) 2018 CERN
 * Author: Maciej Suminski <maciej.suminski@cern.ch>
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef NANOFIP_H
#define NANOFIP_H

#include <libmfipp.h>

#include <functional>
#include <mutex>

namespace std {
template<>
struct hash<mstrfip_data_status>
{
    size_t operator()(const mstrfip_data_status& a) const noexcept
    {
        // no need for any fancy algorithm, enum values are already unique
        return a;
    }
};
} // namespace std

/**
 * NanoFIP is a base class for all devices taking advantage of nanoFIP agent.
 * It might be also used as a generic interface to nanoFIP.
 */
class nanoFIP : public AgentBase
{
protected:
    typedef std::lock_guard<std::mutex> mtx_lock;

public:
    ///> Produced/consumed variable event handler signature
    typedef std::function<void(nanoFIP&, mstrfip_data*)> VarCallback;

    ///> Map to store data frame stats
    typedef std::unordered_map<mstrfip_data_status, unsigned int> DataStatusMap;

    /**
     * nanoFIP constructor.
     * @param addr is the agent address.
     * @param variable_size is the produced/consumed variable size (in bytes).
     */
    nanoFIP(unsigned int addr, unsigned int variable_size);
    virtual ~nanoFIP() {}

    /**
     * Registers variables in a periodic variable window.
     * @param window is the target window.
     */
    int AddVariables(WindowPeriodicVar& window);

    /**
     * Get consumed variable data.
     * @param idx is the variable byte index.
     * @return Consumed variable data. Note it is fetched from a local cache,
     * and may not always represent the nanoFIP memory contents.
     */
    uint8_t GetConsVarData(unsigned int idx)
    {
        auto lock = getConsLock();
        return getConsDataLock(idx, lock);
    }

    /**
     * Get produced variable data.
     * @param idx is the variable byte index.
     * @return Produced variable data. Note it is fetched from a local cache,
     * and may not always represent the nanoFIP memory contents.
     */
    uint8_t GetProdVarData(unsigned int idx)
    {
        auto lock = getProdLock();
        return getProdDataLock(idx, lock);
    }

    /**
     * Set new produced variable data.
     * @param idx is the variable byte index.
     * @param val is the new value.
     */
    void SetProdVarData(unsigned int idx, uint8_t val)
    {
        auto lock = getProdLock();
        setProdDataLock(idx, val, lock);
    }

    ///> Return produced/consumed variable size (their size is equal in nanoFIP)
    size_t GetVarSize() const
    {
        return cons_data.size();
    }

    ///> Set user-defined callback called after the consumed variable is updated
    void SetConsVarCb(VarCallback cb)
    {
        cons_cb = cb;
    }

    ///> Set user-defined callback called after the produced variable is updated
    void SetProdVarCb(VarCallback cb)
    {
        prod_cb = cb;
    }

    ///> Returns frame statistics
    const DataStatusMap& GetDataStats() const
    {
        return data_stats;
    }

    ///> Resets frame statistics
    void ClearDataStats();

    ///> Returns human-readable description for data status enum
    static const std::string& GetDataStatDesc(mstrfip_data_status stat);

protected:
    ///> Consumed variable event handler
    virtual void consVarHandler(mstrfip_dev *dev, mstrfip_data *cvar,
            mstrfip_irq* irq);

    ///> Produced variable event handler
    virtual void prodVarHandler(mstrfip_dev *dev, mstrfip_data *pvar,
            mstrfip_irq* irq);

    ///> nanoFIP produced variable ID
    static const uint16_t PROD_VAR_ID = 0x0005;

    ///> nanoFIP consumed variable ID
    static const uint16_t CONS_VAR_ID = 0x0006;

    ///> Used-defined consumed variable callback (called by consVarHandler)
    VarCallback cons_cb;

    ///> User-defined produced variable callback (called by prodVarHandler)
    VarCallback prod_cb;

    ///> Acquires a lock for consumed variable
    std::unique_ptr<mtx_lock> getConsLock()
    {
        return std::unique_ptr<mtx_lock>(new mtx_lock(cons_mtx));
    }

    ///> Acquires a lock for produced variable
    std::unique_ptr<mtx_lock> getProdLock()
    {
        return std::unique_ptr<mtx_lock>(new mtx_lock(prod_mtx));
    }

    ///> Method to read consumed variable data when a lock is acquired
    uint8_t getConsDataLock(unsigned int idx, std::unique_ptr<mtx_lock>&)
    {
        return cons_data[idx];
    }

    ///> Method to read produced variable data when a lock is acquired
    uint8_t getProdDataLock(unsigned int idx, std::unique_ptr<mtx_lock>&)
    {
        return prod_data[idx];
    }

    ///> Method to write produced variable data when a lock is acquired
    void setProdDataLock(unsigned int idx, uint8_t val, std::unique_ptr<mtx_lock>&)
    {
        prod_data[idx] = val;
        prod_dirty = true;
    }

private:
    ///> Data frame stats
    DataStatusMap data_stats;

    ///> Raw consumed variable data (updated by consVarHandler), guarded with cons_mtx
    std::vector<uint8_t> cons_data;

    ///> Raw produced variable data (used by prodVarHandler), guarded with prod_mtx
    std::vector<uint8_t> prod_data;

    ///> Flag indicating that production variable has been modified and
    ///> needs to be sent to nanoFIP
    bool prod_dirty;

    ///> Mutex guarding the consumed variable data
    std::mutex cons_mtx;

    ///> Mutex guarding the produced variable data
    std::mutex prod_mtx;
};

#endif /* NANOFIP_H */
